;;; Mudsync --- Live hackable MUD
;;; Copyright © 2016 Christopher Allan Webber <cwebber@dustycloud.org>
;;;
;;; This file is part of Mudsync.
;;;
;;; Mudsync is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Mudsync is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Mudsync.  If not, see <http://www.gnu.org/licenses/>.

(define-module (mudsync networking)
  #:use-module (8sync systems actors)
  #:use-module (8sync agenda)
  #:use-module (ice-9 format)
  #:use-module (ice-9 match)
  #:use-module (ice-9 rdelim)
  #:use-module (oop goops)

  #:export (;; Should we be exporting these?
            %default-server
            %default-port

            <network-manager>
            nm-close-everything))

;;; Networking
;;; ==========

(define %default-server #f)
(define %default-port 8889)

(define-class <network-manager> (<actor>)
  (server-socket #:getter nm-server-socket)
  ;; mapping of client -> client-id
  (clients #:getter nm-clients
           #:init-thunk make-hash-table)
  ;; send input to this actor
  (send-input-to #:getter nm-send-input-to
                 #:init-keyword #:send-input-to)
  (message-handler
   #:init-value
   (make-action-dispatch
    (start-listening
     (lambda* (actor message
                     #:key (server %default-server)
                     (port %default-port))
       (nm-install-socket actor server port)))
    (send-to-client
     (lambda* (actor message #:key client data)
       (nm-send-to-client-id actor client data))))))

;;; TODO: We should provide something like this, but this isn't used currently,
;;;    and uses old deprecated code (the 8sync-port-remove stuff).
;; (define-method (nm-close-everything (nm <network-manager>) remove-from-agenda)
;;   "Shut it down!"
;;   ;; close all clients
;;   (hash-for-each
;;    (lambda (_ client)
;;      (close client)
;;      (if remove-from-agenda
;;          (8sync-port-remove client)))
;;    (nm-clients nm))
;;   ;; reset the clients list
;;   (set! (nm-clients) (make-hash-table))
;;   ;; close the server
;;   (close (nm-server-socket nm))
;;   (if remove-from-agenda
;;       (8sync-port-remove (nm-server-socket nm))))

;; Maximum number of backlogged connections when we listen
(define %maximum-backlog-conns 128)     ; same as SOMAXCONN on Linux 2.X,
                                        ; says the intarwebs

(define (nm-install-socket nm server port)
  "Install socket on SERVER with PORT"
  (define s
    (socket PF_INET  ; ipv4
            SOCK_STREAM  ; two-way connection-based byte stream
            0))
  (define addr
    (if server
        (inet-pton AF_INET server)
        INADDR_LOOPBACK))

  ;; Totally mimed from the Guile manual.  Not sure if we need this, but:
  ;; http://www.unixguide.net/network/socketfaq/4.5.shtml
  (setsockopt s SOL_SOCKET SO_REUSEADDR 1) ; reuse port even if port is busy
  ;; Connecting to a non-specific address:
  ;;   (bind s AF_INET INADDR_ANY port)
  ;; Should this be an option?  Guess I don't know why we'd need it
  ;; @@: If we wanted to support listening on a particular hostname,
  ;;   could see 8sync's irc.scm...
  (bind s AF_INET addr port)
  ;; Listen to connections
  (listen s %maximum-backlog-conns)

  ;; Make port non-blocking
  (fcntl s F_SETFL (logior O_NONBLOCK (fcntl s F_GETFL)))

  ;; @@: This is used in Guile's http server under the commit:
  ;;       * module/web/server/http.scm (http-open): Ignore SIGPIPE. Keeps the
  ;;         server from dying in some circumstances.
  ;;   (sigaction SIGPIPE SIG_IGN)
  ;; Will this break other things that use pipes for us though?

  (slot-set! nm 'server-socket s)

  (format #t "Listening for clients in pid: ~s\n" (getpid))

  ;; TODO: set up periodic close of idle connections?
  (let loop ()
    ;; (yield)  ;; @@: Do we need this?
    (define client-connection (accept s))
    (8sync (nm-new-client nm s client-connection))
    (loop)))

(define (nm-new-client nm s client-connection)
  "Handle new client coming in to socket S"
  (define client-details (cdr client-connection))
  (define client (car client-connection))
  (define client-id (big-random-number))
  (format #t "New client: ~s\n" client-details)
  (format #t "Client address: ~s\n"
          (gethostbyaddr
           (sockaddr:addr client-details)))
  (fcntl client F_SETFL (logior O_NONBLOCK (fcntl client F_GETFL)))
  (hash-set! (nm-clients nm) client-id client)
  (<- nm (nm-send-input-to nm) 'new-client #:client client-id)
  (nm-client-receive-loop nm client client-id))

(define (nm-client-receive-loop nm client client-id)
  "Make a method to receive client data"
  (define (loop)
    (define line (read-line client))
    (if (eof-object? line)
        (nm-handle-port-eof nm client client-id)
        (begin
          (nm-handle-line nm client client-id
                          (string-trim-right line #\return))
          (loop))))
  (loop))

(define (nm-handle-port-closed nm client client-id)
  "Handle a closed port"
  (format #t "DEBUG: handled closed port ~x\n" client-id)
  (hash-remove! (nm-clients nm) client-id)
  (<- nm (nm-send-input-to nm) 'client-closed #:client client-id))

(define-method (nm-handle-port-eof nm client client-id)
  "Handle seeing an EOF on port"
  (format #t "DEBUG: handled eof-object on port ~x\n" client-id)
      (close client)
  (hash-remove! (nm-clients nm) client-id)
  (<- nm (nm-send-input-to nm) 'client-closed #:client client-id))

(define-method (nm-handle-line nm client client-id line)
  "Handle an incoming line of input from a client"
  (<- nm (nm-send-input-to nm) 'client-input
      #:data line
      #:client client-id))

(define-method (nm-send-to-client-id nm client-id data)
  "Send DATA to TO-CLIENT id"
  (define client-obj (hash-ref (nm-clients nm) client-id))
  (if (not client-obj)
      (throw 'no-such-client
             "Asked to send data to client but that client doesn't exist"
             #:client-id client-id
             #:data data))
  (display data client-obj))
